/*
 * Copyright (c) CERN 2013-2015
 *
 * Copyright (c) Members of the EMI Collaboration. 2011-2013
 *  See  http://www.eu-emi.eu/partners for details on the copyright
 *  holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <apr_strings.h>
#include <ctype.h>
#include <dmlite/c/catalog.h>
#include <dmlite/c/checksums.h>
#include <httpd.h>
#include <regex.h>
#include <shared/utils.h>
#include <http_log.h>
#include "mod_lcgdm_disk.h"

#define RFC3230_REGEXP "^([[:alnum:]]+)(;q=[[:digit:]]*)?([,]?)"

static char* dav_disk_strcpy_lower(char *out, const char *in, size_t max)
{
    size_t i;
    for (i = 0; in[i] && i < max; ++i)
        out[i] = tolower(in[i]);
    return out;
}

/**
 * Parses p in order to find the next digest algorithm
 * We don't care much about priorities, so give it as we find them
 * @return 0 when there are no more digest available
 */
int dav_disk_next_digest(const char **want_digest, char *output, size_t outsize)
{
    (void) outsize;

    regex_t regex;
    regmatch_t matches[3];

    if (regcomp(&regex, RFC3230_REGEXP, REG_EXTENDED) != 0)
        abort();

    while (isspace(**want_digest))
        ++(*want_digest);

    if (regexec(&regex, *want_digest, 3, matches, 0) == 0) {
        size_t digest_len = matches[1].rm_eo - matches[1].rm_so;
        const char *digest = (*want_digest) + matches[1].rm_so;

        dav_disk_strcpy_lower(output, digest, digest_len);
        output[digest_len] = '\0';

        *want_digest += matches[0].rm_eo;

        return 1;
    }

    return 0;
}


dav_error* dav_disk_digest_header(request_rec *r, const dav_resource *resource,
        char* output, size_t outsize)
{
    const char *want_digest;
    char digest_name[32], digest[64], full_digest_name[64];
    int has_digest;

    want_digest = apr_table_get(r->headers_in, "Want-Digest");

    if (!resource->info->fd || !want_digest)
        return 0;

    has_digest = 0;
    while (dav_disk_next_digest(&want_digest, digest_name, sizeof(digest_name))) {
        snprintf(full_digest_name, sizeof(full_digest_name), "checksum.%s", digest_name);

        int ret = dmlite_getchecksum(resource->info->ctx,
            resource->info->namespace_path, // logical name
            full_digest_name, digest, sizeof(digest),
            resource->info->loc.chunks[0].url.path, // replica
            0, // 0 = do not recalculate
            5 // 5 seconds wait, otherwise return EAGAIN
            );

        if (ret == 0 && digest[0] != '\0') {
            size_t len;
            len = snprintf(output, outsize, "%s=%s,", digest_name, digest);
            output += len;
            outsize -= len;
            has_digest = 1;
            dmlite_fseek(resource->info->fd, 0, SEEK_SET);
        } else if (DMLITE_ERRNO(dmlite_errno(resource->info->ctx)) == EAGAIN) {
            return dav_shared_new_error(resource->info->request, resource->info->ctx,
                202, "Checksum not available yet. Try again later.");
        } else if (dmlite_errno(resource->info->ctx) != 0) {
            ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, resource->info->request,
                "Failed to get the checksum %s: %s", digest_name, dmlite_error(resource->info->ctx));
        } else {
            ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, resource->info->request,
                "Failed to get the checksum %s: empty value", digest_name);
        }
    }

    if (*(output-1) == ',') {
        --output;
    }
    *output = '\0';

    return NULL;
}
