/*
 * Copyright (c) CERN 2013-2015
 *
 * Copyright (c) Members of the EMI Collaboration. 2011-2013
 *  See  http://www.eu-emi.eu/partners for details on the copyright
 *  holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <apr_file_io.h>
#include <apr_strings.h>
#include <ctype.h>
#include <dmlite/c/dmlite.h>
#include <dmlite/c/catalog.h>
#include <dmlite/c/pool.h>
#include <dmlite/c/utils.h>
#include <httpd.h>
#include <http_log.h>
#include <http_protocol.h>
#include <http_request.h>
#include <inttypes.h>
#include <sys/types.h>

#include "delivery.h"
#include "mime.h"
#include "mod_lcgdm_ns.h"
#include "../shared/config.h"
#include "../shared/utils.h"

/* Used for internal recursivity */
#define DAV_WALKTYPE_POSTFIX 0x8000
#define DAV_CALLTYPE_POSTFIX 1000

/* Advanced declarations */
static dav_error *dav_ns_walk(const dav_walk_params *params, int depth,
        dav_response **response);

/** Structure used for the walking */
struct dav_ns_walker_context
{
    const dav_walk_params *params;

    dav_walk_resource walk_resource;

    dav_buffer sfn;
    dav_buffer uri;
};
typedef struct dav_ns_walker_context dav_ns_walker_context;

/** Structure for streams (dummy, since only empty creates will be handled) */
struct dav_stream
{
};

/**
 * Just iterate the query, and pass the key/values to the dmlite stack
 */
static int dav_ns_pass_query_callback(void *rec, const char *key,
        const char *value)
{
    dmlite_context* ctx = (dmlite_context*)rec;
    dmlite_any* any_value = dmlite_any_new_string(value);
    dmlite_set(ctx, key, any_value);
    dmlite_any_free(any_value);
    return 1;
}

static void dav_ns_pass_query_to_dmlite(apr_pool_t* pool, const apr_table_t* query,
        dmlite_context* ctx)
{
    apr_table_do(dav_ns_pass_query_callback, ctx, query, NULL);
}

/**
 * Append the nameserver to the URL, so the disk can do callbacks
 */
static void dav_ns_append_self(dav_resource_private *info,
        dmlite_location *location)
{
    dmlite_any* self = dmlite_any_new_string(info->sfn);
    dmlite_any_dict_insert(location->chunks[0].url.query, "dav_sfn", self);
    dmlite_any_free(self);
}

/**
 * Append the user who requested the replica
 */
static void dav_ns_append_user(dav_resource_private *info,
        dmlite_location *location)
{
    if (info->user_creds) {
        dmlite_any* user_dn = dmlite_any_new_string(info->user_creds->client_name);
        dmlite_any_dict_insert(location->chunks[0].url.query, "dav_user", user_dn);
        dmlite_any_free(user_dn);
    }
}

/**
 * Gets the location when the server is a DPM
 */
static dav_error *dav_ns_get_dpm_location(apr_pool_t *pool,
        dav_resource_private *info, char force_secure_redirect)
{
    dmlite_location *location = dmlite_get(info->ctx, info->sfn);

    if (location == NULL) {
        return dav_shared_new_error(info->request, info->ctx, 0,
                "Could not get the location for %s", info->sfn);
    }

    if (location->chunks[0].url.domain[0] == '\0') {
        info->virtual_location = location;
        info->is_virtual = 1;
        return NULL;
    }

    /* Append the callback, so the disk can call the nameserver */
    dav_ns_append_self(info, location);

    /* Append the user */
    dav_ns_append_user(info, location);

    /* Destination URL */
    info->redirect = dav_shared_build_url(pool, &location->chunks[0].url,
            &info->d_conf->redirect, force_secure_redirect);
    dmlite_location_free(location);

    return NULL ;
}

/**
 * Gets the location when the server is an LFC
 */
static dav_error *dav_ns_get_lfc_location(apr_pool_t *pool, dav_resource_private *info,
        char force_secure_redirect)
{
    unsigned nReplicas = 0;
    dmlite_replica* replicas = NULL;

    if (dmlite_getreplicas(info->ctx, info->sfn, &nReplicas, &replicas) != 0) {
        return dav_shared_new_error(info->request, info->ctx, 0,
                "Could not get the replicas for %s", info->sfn);
    }
    else if (nReplicas == 0) {
        return dav_shared_new_error(info->request, info->ctx, HTTP_NOT_FOUND,
                "No replicas found for %s", info->sfn);
    }

    /* Destination URL. Should be in order, so pick first */
    dmlite_url* url = dmlite_parse_url(replicas[0].rfn);
    info->redirect = dav_shared_build_url(pool, url,
            &info->d_conf->redirect, force_secure_redirect);

    dmlite_url_free(url);
    dmlite_replicas_free(nReplicas, replicas);

    return NULL ;
}

/**
 * Gets the location when the server is a Plain server
 */
static dav_error *dav_ns_get_plain_location(apr_pool_t *pool,
                                          dav_resource_private *info, char force_secure_redirect)
{
    dmlite_location *location = dmlite_get(info->ctx, info->sfn);

    if (location == NULL) {
        return dav_shared_new_error(info->request, info->ctx, 0,
                                    "Could not get the location for %s", info->sfn);
    }

    if (location->chunks[0].url.domain[0] == '\0') {
        info->virtual_location = location;
        info->is_virtual = 1;
        return NULL;
    }

    /* Destination URL */
    info->redirect = dav_shared_build_url(pool, &location->chunks[0].url,
                                          &info->d_conf->redirect, force_secure_redirect);
    dmlite_location_free(location);

    return NULL ;
}

/**
 * Get the location of the file
 */
dav_error *dav_ns_get_location(apr_pool_t *pool, dav_resource_private *info,
        char force_secure_redirect)
{
    dav_error *nested_error;

    /* Logic here depends on the node type */
    switch (info->s_conf->type) {
        case DAV_NS_NODE_HEAD:
            nested_error = dav_ns_get_dpm_location(pool, info,
                    force_secure_redirect);
            break;

            /* LFC wants the whole list */
        case DAV_NS_NODE_LFC:
            nested_error = dav_ns_get_lfc_location(pool, info,
                    force_secure_redirect);
            break;

        case DAV_NS_NODE_PLAIN:
            nested_error = dav_ns_get_plain_location(pool, info,
                    force_secure_redirect);
            break;

        default:
            return dav_shared_new_error(info->request, NULL,
                    HTTP_INTERNAL_SERVER_ERROR, "Invalid node type configured");
    }

    if (nested_error)
        return nested_error;

    return NULL ;
}

/**
 * Return the parent path of "path"
 * @param pool Where to allocate space
 * @param path The path
 * @return     The parent path
 */
static char *dav_ns_dirname(apr_pool_t *pool, const char *path)
{
    char *parent;
    int i;

    parent = apr_pstrcat(pool, path, NULL );

    /* We skip the last character so if path is a directory, we are fine */
    for (i = strlen(parent) - 2; i >= 0; --i) {
        if (parent[i] == '/') {
            parent[i + 1] = '\0'; /* Keep the slash */
            break;
        }
    }

    return parent;
}

/**
 * Get the context from the connection notes, or creates and initializes.
 * @return NULL on success.
 */
static dav_error *dav_ns_initialize_dmlite_context(dav_resource_private *info)
{
    info->ctx = (dmlite_context*) apr_table_get(
            info->request->connection->notes, info->d_conf->context_key);

    if (info->ctx == NULL ) {
        dmlite_manager* manager;
        if (info->d_conf->manager)
            manager = info->d_conf->manager;
        else if (info->s_conf->manager)
            manager = info->s_conf->manager;
        else
            return dav_shared_new_error(info->request, NULL, HTTP_INTERNAL_SERVER_ERROR,
                    "No dmlite manager found. Probably the server is misconfigured.");

        info->ctx = dmlite_context_new(manager);
        if (info->ctx == NULL) {
            return dav_shared_new_error(info->request, NULL,
                    HTTP_INTERNAL_SERVER_ERROR,
                    "Could not instantiate a context: %s",
                    dmlite_manager_error(manager));
        }
        dmlite_any* protocol;
        if (is_ssl_used(info->request))
            protocol = dmlite_any_new_string("https");
        else
            protocol = dmlite_any_new_string("http");
        dmlite_set(info->ctx, "protocol", protocol);
        dmlite_any_free(protocol);

        apr_pool_pre_cleanup_register(info->request->connection->pool,
                info->ctx, dav_shared_context_free);

        if (!(info->d_conf->flags & DAV_NS_NOAUTHN)) {
            apr_pool_t *subpool;

            apr_pool_create(&subpool, info->request->pool);
            info->user_creds = dav_shared_get_user_credentials(subpool, info->request,
                    info->d_conf->anon_user, info->d_conf->anon_group,
                    info->d_conf->trusted_dns);

            if (!info->user_creds) {
                return dav_shared_new_error(info->request, NULL, HTTP_FORBIDDEN,
                        "Can not authenticate the user");
            }

            if (dmlite_setcredentials(info->ctx, info->user_creds) != 0) {
                return dav_shared_new_error(info->request, info->ctx,
                        HTTP_FORBIDDEN, "Could not set credentials");
            }

            apr_pool_destroy(subpool);
        }

        apr_table_setn(info->request->connection->notes, info->d_conf->context_key,
                (char*) info->ctx);
    }
    else {
        info->user_creds = dmlite_get_security_context(info->ctx)->credentials;
    }
    return NULL ;
}

/**
 * This function is used internally to create new resources.
 * @param r          The request to associate.
 * @param sfn        The SFN to link to the resource.
 * @param uri        The URI (without the query!)
 * @param user       The associated user. If NULL, will be queried and initialized.
 * @param resource   Where to put it.
 * @return           NULL on success.
 */
static dav_error *dav_ns_internal_get_resource(request_rec *r, const char *sfn,
        dav_resource **resource)
{
    dav_resource_private *info;
    int exists;
    apr_table_t *query;
    unsigned nargs;
    dav_error *nested_error;

    /* Initialize info */
    info = apr_pcalloc(r->pool, sizeof(dav_resource_private));
    info->stat.extra = dmlite_any_dict_new();
    apr_pool_pre_cleanup_register(r->pool, info->stat.extra,
            dav_shared_dict_free);

    info->request = r;
    info->sfn = sfn;

    info->s_conf = ap_get_module_config(r->server->module_config, &lcgdm_ns_module) ;
    info->d_conf = ap_get_module_config(r->per_dir_config, &lcgdm_ns_module) ;

    /* Instantiate the catalog, if not in the connection notes */
    nested_error = dav_ns_initialize_dmlite_context(info);
    if (nested_error)
        return nested_error;

    /* Extract parameters from the query */
    query = dav_shared_parse_query(r->pool, r->parsed_uri.query, &nargs);

#ifdef WITH_METALINK
    /* Explicit */
    info->metalink = (apr_table_get(query, "metalink") != NULL );
    /* Through headers maybe? (i.e. aria2c) */
    if (!info->metalink && dav_shared_request_accepts(r, "application/metalink+xml")) {
        info->metalink = 1;
    }
#else
    info->metalink = 0;
#endif

    /* Check if the user wants the new UI */
    const char* cookie_str = apr_table_get(r->headers_in, "Cookie");
    if (cookie_str) {
        apr_table_t* cookies = dav_lcgdm_parse_cookies(r->pool, cookie_str);
        const char* ui_cookie = apr_table_get(cookies, "lcgdm_dav.ui");
        info->new_ui = (ui_cookie && strncmp(ui_cookie, "new", 3) == 0);
    }

    /* Check if the user wants secure redirection */
    const char *secure_redirection = apr_table_get(r->headers_in, "Secure-Redirection");
    info->force_secure = (secure_redirection && strncmp(secure_redirection, "0", 1) != 0);

    /* Pass everything to dmlite */
    dav_ns_pass_query_to_dmlite(r->pool, query, info->ctx);

    /* Stat */
    switch (dmlite_statx(info->ctx, info->sfn, &(info->stat))) {
        /* We are good */
        case 0:
            exists = 1;
            break;

        /* Does not exists */
        case ENOENT:
            exists = 0;
            /* When creating a file, break here */
            if (r->method_number == M_PUT
                    || r->method_number == M_MKCOL
                    || r->method_number == M_MOVE
                    || r->method_number == M_OPTIONS
                    || r->method_number == M_COPY)
                break;
            /* Not creating and not exists, fail */
            return dav_shared_new_error(r, info->ctx, 0, "%s does not exist",
                                info->sfn);

        /* Access denied */
        case EACCES:
            /* If this is not a HEAD or a GET (both are M_GET), just fail,
             * but keep going otherwise */
            if (r->method_number != M_GET)
                return dav_shared_new_error(r, info->ctx, 0,
                        "Access forbidden for %s on %s", info->sfn, r->method);
            /* no break */

        default:
            return dav_shared_new_error(r, info->ctx, 0, "Can not stat %s",
                    info->sfn);
    }

    /* Create and return resource */
    *resource = apr_pcalloc(r->pool, sizeof(dav_resource));
    (*resource)->type = DAV_RESOURCE_TYPE_REGULAR;
    (*resource)->exists = exists;
    (*resource)->collection = ((info->stat.stat.st_mode & S_IFDIR) != 0);
    (*resource)->uri = sfn;
    (*resource)->info = info;
    (*resource)->hooks = &dav_ns_hooks_repository;
    (*resource)->pool = r->pool;

    if (exists)
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                "Resource for %s (%s) found (dir=%d)", (*resource)->uri,
                (*resource)->info->sfn, (*resource)->collection);
    else
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                "NULL resource for %s (%s) created", (*resource)->uri,
                (*resource)->info->sfn);

    return NULL ;
}

/**
 * Generate the resource used for later operations
 * First security checkings can be done here (like user mapping)
 * @param r                The request coming from Apache
 * @param root_dir         The root dir of the request (as configured in
 *                         Apache with Location)
 * @param label            Used with version control. Not relevant.
 * @param use_checked_in   Used with version control. Not relevant.
 * @param resource         Where to put the generated resource
 * @return                 NULL on success. A dav_error instance if not.
 */
static dav_error *dav_ns_get_resource(request_rec *r, const char *root_dir,
        const char *label, int use_checked_in, dav_resource **resource)
{
    (void) label;
    (void) use_checked_in;

    char *sfn;
    dav_error *err;
    int len;

    /* Compose path */
    len = strlen(root_dir);
    if (root_dir[len - 1] == '/') {
        char *tmp = apr_pstrdup(r->pool, root_dir);
        --len;
        tmp[len] = '\0';
        root_dir = tmp;
    }

    sfn = r->parsed_uri.path ? r->parsed_uri.path : "";

    /* Create resource */
    err = dav_ns_internal_get_resource(r, sfn, resource);
    if (err)
        return err;

    /* Directories MUST finish with '/', so redirect if not */
    len = strlen(sfn);
    if ((*resource)->collection && (len == 0 || sfn[len - 1] != '/')) {
        (*resource)->info->redirect = apr_pstrcat(r->pool,
                ap_escape_uri(r->pool, sfn), "/", r->args ? "?" : "",
                r->args ? r->args : "", NULL );
        (*resource)->uri = apr_pstrcat(r->pool, (*resource)->uri, "/", NULL );
        (*resource)->info->sfn = apr_pstrcat(r->pool, (*resource)->info->sfn,
                "/", NULL );
    }

    /* If resource does not exists, or it is a file, it can be copied to/from a remote
     * endpoint
     */
    if (!(*resource)->exists ||
        (*resource)->type == DAV_RESOURCE_TYPE_REGULAR) {
        dav_lcgdm_notify_support_external_copy(r);
    }

    /* We are done here */
    return NULL ;
}

/**
 * Get the parent of a resource (if any)
 * @param resource        The resource to process
 * @param parent_resource The parent, if any
 * @return                NULL on success
 */
static dav_error *dav_ns_get_parent(const dav_resource *resource,
        dav_resource **parent_resource)
{
    char *sfn;
    request_rec *dup_rec;

    /* The parent of the root is itself (kinda) */
    if (strcmp(resource->info->sfn, "/") == 0) {
        *parent_resource = (dav_resource*) resource;
        return NULL ;
    }

    /* Strip last element of the path */
    sfn = dav_ns_dirname(resource->pool, resource->info->sfn);

    /* Duplicate request, removing query from the parsed URI.
     * This is because on PUT requests with query string, it will remain
     * in the parent, making get_resource to behave like it does not exist :-(
     */
    dup_rec = apr_pcalloc(resource->pool, sizeof(request_rec));
    *dup_rec = *resource->info->request;
    dup_rec->parsed_uri.query = NULL;

    /* Return whatever the internal says */
    return dav_ns_internal_get_resource(dup_rec, sfn, parent_resource);
}

/**
 * Compare two resources to check if they are the same
 * @param res1 Resource 1
 * @param res2 Resource 2
 * @return     1 if they are the same
 */
static int dav_ns_is_same(const dav_resource *res1, const dav_resource *res2)
{
    const char *destination;
    apr_uri_t uri1, uri2;

    /* Parse resource 1 */
    apr_uri_parse(res1->pool, res1->uri, &uri1);

    /* Parse resource 2 */
    destination = apr_table_get(res2->info->request->headers_in, "destination");
    if (destination)
        apr_uri_parse(res2->pool, destination, &uri2);
    else
        apr_uri_parse(res2->pool, res2->uri, &uri2);

    /* Host may not appear */
    if (!uri1.hostname)
        uri1.hostname = res1->info->request->server->server_hostname;
    if (!uri2.hostname)
        uri2.hostname = res2->info->request->server->server_hostname;

    /* Compare */
    return strcmp(uri1.hostname, uri2.hostname) == 0
            && strcmp(uri1.path, uri2.path) == 0;
}

/**
 * Check if a given resource is the parent of another one
 * @param res1 Parent to check
 * @param res2 Child
 * @return     1 if res1 is the parent of res2
 */
static int dav_ns_is_parent(const dav_resource *res1, const dav_resource *res2)
{
    char *res2_parent_sfn;

    res2_parent_sfn = dav_ns_dirname(res2->pool, res2->info->sfn);

    return strcmp(res1->info->sfn, res2_parent_sfn) == 0;
}

/**
 * Open a stream. Used for PUT
 * @param resource The resource to open
 * @param mode     The mode (basically, truncate or random, which we don't support)
 * @param stream   Where to put the created structure
 * @return         NULL on success
 */
static dav_error *dav_ns_open_stream(const dav_resource *resource,
        dav_stream_mode mode, dav_stream **stream)
{
    (void) mode;

    dav_resource_private *info;
    unsigned has_range;
    const char *range, *length;
    off64_t contentLength = 0;
    int errCode;

    info = resource->info;

    /* Must be writable (if we are here, they want to write) */
    if (!(info->d_conf->flags & DAV_NS_WRITE))
        return dav_shared_new_error(info->request, NULL, HTTP_FORBIDDEN,
                "Configured as read-only endpoint (%s)", resource->uri);

    /* If content-length is 0, call create instead of put! */
    length = apr_table_get(info->request->headers_in, "content-length");
    if (length != NULL ) {
        contentLength = atol(length);
        if (contentLength == 0) {
            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, info->request,
                    "PUT with content-length 0. Creating ns entry.");

            if (dmlite_create(info->ctx, info->sfn, 0644) != 0)
                return dav_shared_new_error(info->request, info->ctx, 0,
                        "Could not create empty file %s", info->sfn);

            *stream = (dav_stream*) calloc(1, sizeof(dav_stream));

            return NULL ;
        }
    }

    /* NS alone doesn't support PUTs with content! */
    if (info->s_conf->type == DAV_NS_NODE_LFC) {
        return dav_shared_new_error(info->request, NULL, HTTP_NOT_IMPLEMENTED,
                "LFC does not support PUTs");
    }

    /*  Range header? */
    range = apr_table_get(info->request->headers_in, "content-range");
    has_range = (range != NULL );
    if (has_range) {
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, info->request, "Range: %s",
                range);
    }

    /* Pass expected content-length if available */
    if (contentLength) {
        dmlite_any* any = dmlite_any_new_u64(contentLength);
        errCode = dmlite_set(info->ctx, "requested_size", any);
        dmlite_any_free(any);
        if (errCode)
            ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, info->request,
                    "Tried to set the requested size, but failed");
        else
            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, info->request,
                    "Set requested_size to %ld", contentLength);
    }

    /* Call dmlite_put to see where to go */
    dmlite_location *location;
    location = dmlite_put(info->ctx, info->sfn);

    errCode = dmlite_errno(info->ctx);
    int httpCode = 0;
    if (errCode == EINVAL)
        httpCode = HTTP_BAD_REQUEST;
    if (errCode != 0) {
        return dav_shared_new_error(info->request, info->ctx, httpCode, "Can not put %s",
                                    info->sfn);
    }

    /* Redirect */
    info->redirect = dav_shared_build_url(resource->pool,
            &location->chunks[0].url, &info->d_conf->redirect, info->force_secure);

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, info->request,
            "PUT request to be done in %s", info->redirect);

    apr_table_set(info->request->err_headers_out, "Location", info->redirect);

    dmlite_location_free(location);

    /* To allow web clients to handle themselves the Location header, if
     * X-No-Redirect is present, use 202 Accepted instead of 307 Temporary redirect
     */
    int status_code = HTTP_TEMPORARY_REDIRECT;
    if (apr_table_get(info->request->headers_in, "x-no-redirect")) {
        status_code = HTTP_ACCEPTED;
    }

    /* Standard says 301/302, but some clients will retry in that case with a GET.
     * 307 is unambiguous */
    return dav_shared_new_error(info->request, NULL, status_code,
            "=> %s", info->redirect);
}

/**
 * Close a stream
 * @param stream The stream to close
 * @param commit Commit the change
 * @return       NULL on success
 */
static dav_error *dav_ns_close_stream(dav_stream *stream, int commit)
{
    (void) stream;
    (void) commit;

    /* Nothing to do here */
    return NULL ;
}

/**
 * Write to a stream
 * @param stream  The stream where to write
 * @param buf     The data to write
 * @param bufsize How much data there is in the buffer
 * @return        NULL on success
 */
static dav_error *dav_ns_write_stream(dav_stream *stream, const void *buf,
        apr_size_t bufsize)
{
    (void) stream;
    (void) buf;
    (void) bufsize;
    /* We never enter here (either redirected, either empty PUT's) */
    return NULL ;
}

/**
 * Moves to an specific position within a stream
 * @param stream       The stream to move
 * @param abs_position Where to move
 * @return             NULL on success
 */
static dav_error *dav_ns_seek_stream(dav_stream *stream, apr_off_t abs_position)
{
    (void) stream;
    (void) abs_position;
    /* Nothing to do here */
    return NULL ;
}

/**
 * Sets the HTTP headers, like the size of the content, the modification date...
 * @param r        The request
 * @param resource The resource generated previously
 * @return         NULL on success
 */
static dav_error *dav_ns_set_headers(request_rec *r,
        const dav_resource *resource)
{
    dav_resource_private *info;
    dav_error *err;
    const char *etag;
    char buffer[1024];

    /* Null resource */
    if (!resource->exists)
        return NULL ;

    info = resource->info;

    /* ETag */
    etag = resource->hooks->getetag(resource);
    apr_table_setn(r->headers_out, "ETag", etag);

    /* NS saves the time in seconds since epoch, while Apache expects microseconds */
    ap_update_mtime(r, info->stat.stat.st_mtime * 1000000);
    ap_set_last_modified(r);

    if (!resource->collection && !resource->info->metalink && !resource->info->is_virtual) {
        if ((err = dav_ns_get_location(r->pool, info, info->force_secure)))
            return err;

        if (!info->is_virtual) {
            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                    "Resource %s located at %s", info->sfn, info->redirect);

            apr_table_setn(r->headers_out, "Location", info->redirect);
#ifdef WITH_METALINK
            snprintf(buffer, sizeof(buffer),
                    "<%s?metalink>; rel=describedby; type=\"application/metalink+xml\"",
                    info->request->uri);
            apr_table_set(r->err_headers_out, "Link", buffer);
#endif
            return dav_shared_new_error(r, NULL, HTTP_MOVED_TEMPORARILY, "=> %s",
                    info->redirect);
        }
        else {
            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                    "Resource %s is a virtual file", info->sfn);
        }
    }

    /* Need to redirect? (Missing trailing /) */
    if (info->redirect) {
        apr_table_setn(r->headers_out, "Location", info->redirect);
        return dav_shared_new_error(r, NULL, HTTP_MOVED_PERMANENTLY, "=> %s",
                info->redirect);
    }

    /* Content-type */
    if (resource->info->metalink) {
        // If application/metalink+xml not explicitly accepted, assume it is a browser and
        // give generic application/xml so it can be seen
        if (dav_shared_request_accepts(r, "application/metalink+xml"))
            ap_set_content_type(r, "application/metalink+xml");
        else
            ap_set_content_type(r, "application/xml");

        snprintf(buffer, sizeof(buffer), "inline; filename=\"%s.metalink\"",
                info->stat.name);
        apr_table_set(r->headers_out, "Content-Disposition", buffer);
    }
    else if (resource->collection) {
        ap_set_content_type(r, "text/html");
    }
    else if (resource->info->is_virtual) {
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                    "Resource %s is a virtual file", info->sfn);
    }

    /* Nice thing is, Apache takes care of implementing this */
    apr_table_setn(r->headers_out, "Accept-Ranges", "bytes");

    return NULL ;
}

/**
 * This is the second most important function, as it delivers the content of the
 * file (or the directory if not DAV)
 * @param resource The resource generated previously
 * @param output   Here is where the output must be written
 * @return         NULL on success
 */
static dav_error *dav_ns_deliver(const dav_resource *resource,
        ap_filter_t *output)
{
    apr_bucket_brigade *bb;
    apr_bucket *bkt;
    dav_error *err;

    bb = apr_brigade_create(resource->pool, output->c->bucket_alloc);

    if (resource->collection) {
        err = dav_ns_deliver_collection(resource,
                resource->info->request->output_filters, bb);
    }
    else if (resource->info->metalink) {
        err = dav_ns_deliver_metalink(resource,
                resource->info->request->output_filters, bb);
    }
    else if (resource->info->is_virtual) {
        err = dav_ns_deliver_virtual(resource,
                resource->info->request->output_filters, bb);
    }
    else {
        err = dav_shared_new_error(resource->info->request, NULL,
                HTTP_INTERNAL_SERVER_ERROR,
                "NS should not be trying to deliver files!");
    }

    if (err != NULL )
        return err;

    /* Close and flush the output */
    bkt = apr_bucket_eos_create(output->c->bucket_alloc);
    APR_BRIGADE_INSERT_TAIL(bb, bkt);
    if (ap_pass_brigade(resource->info->request->output_filters,
            bb) != APR_SUCCESS)
        return dav_shared_new_error(resource->info->request, NULL,
                HTTP_INTERNAL_SERVER_ERROR, "Could not write EOS to filter.");

    /* All OK */
    return NULL ;
}

/**
 * Create a new collection (i.e. directory)
 * @param resource The resource to be created
 * @return         NULL on success
 */
static dav_error *dav_ns_create_collection(dav_resource *resource)
{
    /* Must be writable */
    if (!(resource->info->d_conf->flags & DAV_NS_WRITE))
        return dav_shared_new_error(resource->info->request, NULL,
                HTTP_FORBIDDEN, "Configured as read-only endpoint (%s)",
                resource->uri);

    /* Must not exist */
    if (resource->exists)
        return dav_shared_new_error(resource->info->request, 0, HTTP_CONFLICT,
                "The file already exists (%s)", resource->info->sfn);

    /* Create */
    switch (dmlite_mkdir(resource->info->ctx, resource->info->sfn, 0775)) {
        case 0:
            break;
        case ENOENT:
            return dav_shared_new_error(resource->info->request,
                    resource->info->ctx, HTTP_CONFLICT,
                    "Could not create the directory %s", resource->info->sfn);
        default:
            return dav_shared_new_error(resource->info->request,
                    resource->info->ctx, 0, "Could not create the directory %s",
                    resource->info->sfn);
    }

    resource->exists = 1;
    resource->collection = 1;
    if (dmlite_statx(resource->info->ctx, resource->info->sfn,
            &(resource->info->stat)) != 0) {
        /* It has just been created, we shouldn't enter here, but just in case */
        return dav_shared_new_error(resource->info->request,
                resource->info->ctx, 0,
                "dm_xstat failed just after the creation of %s",
                resource->info->sfn);
    }

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, resource->info->request,
            "%s created", resource->info->sfn);

    return NULL ;
}

/**
 * Trigger a remote copy
 * @param src       The source. It must exist
 * @param dst       The destination.
 * @param response  The response to the copy process
 * @return          NULL on success
 */
static dav_error *dav_ns_remote_copy(const dav_resource *src, const char *dst,
        dav_response **response)
{
    (void) dst;
    (void) response;

    if (!(src->info->d_conf->flags & DAV_NS_REMOTE_COPY)) {
        return dav_shared_new_error(src->info->request, NULL,
                HTTP_METHOD_NOT_ALLOWED,
                "This server does not allow remote copies");
    }

    dav_error *error;
    error = dav_ns_get_location(src->info->request->pool, src->info, 1);
    if (error)
        return error;


    apr_table_setn(src->info->request->headers_out, "Location",
            src->info->redirect);

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, src->info->request,
            "COPY of '%s' redirected to '%s'", src->info->sfn,
            src->info->redirect);

    return dav_shared_new_error(src->info->request, NULL,
            HTTP_TEMPORARY_REDIRECT, "=> %s", src->info->redirect);
}

/**
 * Trigger a remote fetch
 * @param src       The remote source.
 * @param dst       The local destination.
 * @param response  The response to the copy process
 * @return          NULL on success
 */
static dav_error* dav_ns_remote_fetch(const char *src,
        const dav_resource *dst, dav_response **response)
{
    if (dst->exists) {
        return dav_shared_new_error(dst->info->request, NULL,
                HTTP_CONFLICT, "File exists %s", dst->uri);
    }

    if (!(dst->info->d_conf->flags & DAV_NS_REMOTE_COPY)) {
        return dav_shared_new_error(dst->info->request, NULL,
                HTTP_METHOD_NOT_ALLOWED,
                "This server does not allow remote fetches");
    }

    // Delegate to the write handler, since it does pretty much what we
    // want to do
    apr_table_unset(dst->info->request->headers_in, "content-range");
    apr_table_unset(dst->info->request->headers_in, "content-length");
    return dav_ns_open_stream(dst, DAV_MODE_WRITE_TRUNC, NULL);
}

/**
 * Local copies not allowed
 * @param src       The source. It must exist
 * @param dst       The destination. Must not exist
 * @param depth     Copy in depth? (Collections)
 * @param response  The response to the copy process
 * @return          NULL on success
 */
static dav_error *dav_ns_copy(const dav_resource *src, dav_resource *dst,
        int depth, dav_response **response)
{
    (void) depth;
    return dav_ns_remote_copy(src, dst->uri, response);
}

/**
 * Move a resource
 * @param src       The source. It must exist
 * @param dst       The destination. It must not exist
 * @param response  The response to the move process
 * @return          NULL on success
 */
static dav_error *dav_ns_move(dav_resource *src, dav_resource *dst,
        dav_response **response)
{
    /* Must be writable */
    if (!(src->info->d_conf->flags & DAV_NS_WRITE))
        return dav_shared_new_error(dst->info->request, 0, HTTP_FORBIDDEN,
                "Configured as read-only endpoint (%s)", dst->uri);

    /* Change the name, that's all */
    if (dmlite_rename(src->info->ctx, src->info->sfn, dst->info->sfn) != 0) {
        return dav_shared_new_error(dst->info->request, src->info->ctx, 0,
                "Could not move from %s to %s", src->info->sfn, dst->info->sfn);
    }

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, src->info->request,
            "Moved %s to %s", src->info->sfn, dst->info->sfn);

    *response = NULL;
    return NULL ;
}

/**
 * Remove a resource from DPM, either file or directory
 * NOT recursive, resource has to be a leaf, or this will fail
 * @param resource The resource to remove
 * @return         NULL on success
 */
static dav_error *dav_ns_internal_remove(const dav_resource *resource)
{
    int error = 0;

    if (resource->collection) {
        /* Remove a directory is straigh-forward */
        if (dmlite_rmdir(resource->info->ctx, resource->info->sfn) != 0) {
            error = 1;
        }
        else {
            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, resource->info->request,
                    "Directory %s removed", resource->info->sfn);
        }
    }
    else {
        /* A file is not much more difficult :-) */
        if (dmlite_unlink(resource->info->ctx, resource->info->sfn) != 0) {
            error = 1;
        }
        else {
            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, resource->info->request,
                    "File %s removed", resource->info->sfn);
        }
    }

    if (error) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, resource->info->request,
                "Could not remove %s (%s)", resource->info->sfn,
                dmlite_error(resource->info->ctx));
        return dav_shared_new_error(resource->info->request,
                resource->info->ctx, 0, "Could not remove %s",
                resource->info->sfn);
    }

    return NULL ;
}

/**
 * Callback function for remove walking
 * @param wres      The walking resource
 * @param calltype  Calltype
 * @return          NULL on success
 */
static dav_error *dav_ns_remove_callback(dav_walk_resource *wres, int calltype)
{
    /* Only remove elements that exist, and empty directories */
    if (wres->resource->exists
            && (!wres->resource->collection || calltype == DAV_CALLTYPE_POSTFIX)) {

        dav_error *err;

        err = dav_ns_internal_remove(wres->resource);

        /* We add to the multistatus, not as return */
        if (err)
            dav_shared_add_response(wres, err);
    }
    return NULL ;
}

/**
 * Remove a resource
 * @param resource The resource to remove
 * @param response The response to the request
 * @return         NULL on success
 */
static dav_error *dav_ns_remove(dav_resource *resource, dav_response **response)
{
    dav_error *err;

    /* Be careful: mod_dav will try to remove the file when copying, if source
     * and dest happen to have the same path (even if the host is different).
     * Just ignore those requests.
     */
    if (resource->info->request->method_number == M_COPY) {
        ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, resource->info->request,
                "Ignoring remove requests on COPY");
        return NULL ;
    }

    /* Must be writable */
    if (!(resource->info->d_conf->flags & DAV_NS_WRITE))
        return dav_shared_new_error(resource->info->request, 0, HTTP_FORBIDDEN,
                "Configured as read-only endpoint (%s)", resource->uri);

    if (resource->collection) {
        /* Recursive if collection */
        dav_walk_params params = {
                0, NULL, NULL, NULL, NULL, NULL };
        dav_response *multi_status;

        params.walk_type = DAV_WALKTYPE_POSTFIX;
        params.func = dav_ns_remove_callback;
        params.pool = resource->pool;
        params.root = resource;

        if ((err = dav_ns_walk(&params, DAV_INFINITY, &multi_status)) != NULL )
            return err;

        if ((*response = multi_status) != NULL ) {
            /* Multistatus */
            return dav_shared_new_error(resource->info->request, NULL,
                    HTTP_MULTI_STATUS, "Errors on multistatus");
        }

        resource->exists = 0;
        resource->collection = 0;

        return NULL ;
    }
    else {
        /* File is pretty easy */
        err = dav_ns_internal_remove(resource);
        if (err)
            return err;
        resource->exists = 0;
        resource->collection = 0;
        return NULL ;
    }
}

/**
 * Recursive function to walk through the filesystem
 * @param ctx The context
 * @param depth The maximum depth
 * @return NULL on success
 */
static dav_error *dav_ns_walker(dav_ns_walker_context *ctx, int depth)
{
    dav_error *err = NULL;
    void *dir;
    dmlite_xstat *entry;
    apr_pool_t *pool;
    const dav_resource *this_res;
    dav_resource reusable;
    apr_pool_t *subpool;

    pool = ctx->params->pool;

    /* Save for later */
    this_res = ctx->walk_resource.resource;

    /* Call the function with the current root */
    err = (*ctx->params->func)(&(ctx->walk_resource),
            ctx->walk_resource.resource->collection ?
                    DAV_CALLTYPE_COLLECTION : DAV_CALLTYPE_MEMBER);
    if (err)
        return err;
    if (depth == 0 || !ctx->walk_resource.resource->collection)
        return NULL ;

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
            ctx->walk_resource.resource->info->request, "Descending into %s",
            ctx->sfn.buf);

    /* Open the directory */
    if ((dir = dmlite_opendir(this_res->info->ctx, ctx->sfn.buf)) == NULL ) {
        dav_shared_add_response(&ctx->walk_resource,
                dav_shared_new_error(this_res->info->request,
                        this_res->info->ctx, 0,
                        "Could not open the directory %s", ctx->sfn.buf));
        return NULL ;
    }

    /* Create the subpool */
    apr_pool_create(&subpool, ctx->walk_resource.pool);

    /* Initialize the reusable resource */
    reusable = *this_res;
    reusable.info = NULL;
    reusable.uri = NULL;
    reusable.info = apr_pcalloc(reusable.pool, sizeof(dav_resource_private));
    reusable.info->request = this_res->info->request;
    reusable.info->ctx = this_res->info->ctx;
    reusable.info->s_conf = this_res->info->s_conf;
    reusable.info->d_conf = this_res->info->d_conf;
    reusable.pool = subpool;

    ctx->walk_resource.resource = &reusable;

    /* Scan the directory */
    while ((entry = dmlite_readdirx(this_res->info->ctx, dir))) {
        apr_size_t uri_prev_len;
        apr_size_t path_prev_len;
        int len;

        apr_pool_clear(subpool);

        /* Append the file to path and URI */
        len = strlen(entry->name);
        dav_buffer_place_mem(pool, &ctx->uri, entry->name, len + 1, 1);
        dav_buffer_place_mem(pool, &ctx->sfn, entry->name, len + 1, 1);

        reusable.info->sfn = ctx->sfn.buf;

        /* dav_buffer_place_mem do NOT adjust len */
        uri_prev_len = ctx->uri.cur_len;
        path_prev_len = ctx->sfn.cur_len;
        ctx->sfn.cur_len += len;
        ctx->uri.cur_len += len;

        if (entry->stat.st_mode & S_IFDIR) {
            ctx->sfn.cur_len++;
            ctx->uri.cur_len++;

            ctx->uri.buf[ctx->uri.cur_len - 1] = ctx->sfn.buf[ctx->sfn.cur_len
                    - 1] = '/';
            ctx->uri.buf[ctx->uri.cur_len] = ctx->sfn.buf[ctx->sfn.cur_len] =
                    '\0';
        }

        /* Set "entry-related" fields */
        reusable.uri = ctx->uri.buf;
        reusable.collection = S_ISDIR(entry->stat.st_mode);

        reusable.info->stat = *entry;

        /* Attach or descend */
        if (!ctx->walk_resource.resource->collection) {
            /* Add file */
            err = ctx->params->func(&(ctx->walk_resource), DAV_CALLTYPE_MEMBER);
            if (err)
                break;
        }
        else {
            /* Recursive into child */
            if ((err = dav_ns_walker(ctx, depth - 1)) != NULL )
                break;
        }

        /* Restore length, so next dav_buffer_place_mem overrites current entry */
        ctx->uri.cur_len = uri_prev_len;
        ctx->sfn.cur_len = path_prev_len;
    }

    /* Release resources */
    if (dmlite_closedir(this_res->info->ctx, dir))
        return dav_shared_new_error(this_res->info->request,
                this_res->info->ctx, 0, "Could not close the directory %s",
                ctx->sfn.buf);

    if (err)
        return err;

    apr_pool_destroy(subpool);

    /* Recover */
    ctx->walk_resource.resource = this_res;
    ctx->uri.buf[ctx->uri.cur_len] = '\0';
    ctx->sfn.buf[ctx->sfn.cur_len] = '\0';

    /* Postfix iteration */
    if (ctx->params->walk_type & DAV_WALKTYPE_POSTFIX) {
        return ctx->params->func(&(ctx->walk_resource), DAV_CALLTYPE_POSTFIX);
    }

    /* There we go */
    return NULL ;
}

/**
 * Walk through the file system
 * @param params    Walk parameters
 * @param depth     In depth?
 * @param response  The response
 * @return          NULL on success
 */
static dav_error *dav_ns_walk(const dav_walk_params *params, int depth,
        dav_response **response)
{
    dav_ns_walker_context ctx;
    dav_error *err;

    memset(&ctx, 0, sizeof(ctx));

    ctx.params = params;

    ctx.walk_resource.walk_ctx = params->walk_ctx;
    ctx.walk_resource.pool = params->pool;
    ctx.walk_resource.resource = params->root;

    dav_buffer_init(params->pool, &ctx.sfn, params->root->info->sfn);
    dav_buffer_init(params->pool, &ctx.uri, params->root->uri);

    err = dav_ns_walker(&ctx, depth);

    *response = ctx.walk_resource.response;

    return err;
}

/**
 * Get an ETag
 * @param resource The resource from where retrieve the tag
 * @return         The tag value
 */
static const char *dav_ns_getetag(const dav_resource *resource)
{
    if (resource->info->new_ui) {
        return apr_psprintf(resource->pool, "%lx-%lx-ui.new",
                        (unsigned long) resource->info->stat.stat.st_ino,
                        (unsigned long) resource->info->stat.stat.st_mtime);
    }
    else {
        return apr_psprintf(resource->pool, "%lx-%lx",
                (unsigned long) resource->info->stat.stat.st_ino,
                (unsigned long) resource->info->stat.stat.st_mtime);
    }
}

/** Hooks */
const dav_hooks_repository dav_ns_hooks_repository = {
    1,
    dav_ns_get_resource,
    dav_ns_get_parent,
    dav_ns_is_same,
    dav_ns_is_parent,
    dav_ns_open_stream,
    dav_ns_close_stream,
    dav_ns_write_stream,
    dav_ns_seek_stream,
    dav_ns_set_headers,
    dav_ns_deliver,
    dav_ns_create_collection,
    dav_ns_copy,
    dav_ns_move,
    dav_ns_remove,
    dav_ns_walk,
    dav_ns_getetag,
    NULL,
    dav_ns_remote_copy,
    dav_ns_remote_fetch
};
